"""Utilities."""

import logging
import os
from pathlib import Path
from functools import lru_cache
from io import BytesIO
from logging import Logger
from typing import Tuple, Optional
from urllib.parse import urlparse

import PIL

from trendhunter.tuples import Image


@lru_cache
def format_console(name: str) -> Logger:
    logger = logging.getLogger(name)
    logger.propagate = False
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "({asctime}) ({name}) ({levelname}) {message}",
        style="{",
        datefmt="%d-%m-%Y %H:%M",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def resize_image(image: Image, pixels: Tuple[int]) -> BytesIO:
    im = PIL.Image.open(BytesIO(image.image))
    image_buf = BytesIO()

    im.thumbnail(pixels, PIL.Image.ANTIALIAS)
    format_console(__name__).info(
        f'Resized image "{image.url}" to {im.size[0]}, {im.size[1]} (width,'
        " height)."
    )

    ext = os.path.splitext(urlparse(image.url).path)[1][1:]
    im.save(image_buf, method=6, quality=100, format=ext)

    return image_buf


def resolve_path(path: Optional[Path] = None) -> Path:
    path = path or Path(".").resolve()
    if path.is_dir():
        path.mkdir(parents=True, exist_ok=True)
        path = path / "template.pptx"
    else:
        path.parent.mkdir(parents=True, exist_ok=True)
    return path



"""Output formatters."""

from typing import List

from pptx import Presentation

from trendhunter.tuples import Article, Context
from trendhunter.utils import (
    format_console,
    resize_image,
    resolve_path,
)


def console(articles: List[Article], context: Context):
    for article in articles:
        format_console(__name__).debug(
            f"""Article ({article.url.rsplit("/")[-1]}) {{
    url: {article.url}
    title: {article.text.title}
    description: {article.text.description[:40]}... (truncated)
    metadata: {{
        id: {article.text.metadata.eid}
        category: {article.text.metadata.cid}
    }}
    image: {{
        url: {article.image.url}
        image: {article.image.image[:40]}... (truncated)
    }}
}}
"""
        )


def slideshow(articles: List[Article], context: Context) -> None:
    path = resolve_path(context.path)

    if path.exists():
        show = Presentation(path)
    else:
        show = Presentation()

    for article in articles:
        slide = show.slides.add_slide(show.slide_layouts[8])
        slide.placeholders[0].text = article.text.title
        slide.placeholders[1].insert_picture(
            resize_image(article.image, context.pixels)
        )
        slide.placeholders[2].text = article.text.description

    show.save(path)
    format_console(__name__).info(
        f'Formatted {len(articles)} articles to "{path}".'
    )



"""Click interface."""

import logging
from pathlib import Path
from typing import Callable, Optional, Tuple

import click
from aiohttp import ClientError
from slugify import slugify

from trendhunter.api import ResourceError, TrendHunterAPI
from trendhunter.formatters import console, slideshow
from trendhunter.taxonomies import PageType
from trendhunter.tuples import Context
from trendhunter.utils import format_console

FORMATTERS = [console, slideshow]


def catch_execute(func: Callable, *args):
    try:
        yield from func(*args)
    except (ResourceError, ClientError) as e:
        format_console(__name__).critical(f"({type(e).__name__}) {e}")


n = click.option(
    "-n",
    "n",
    default=50,
    type=click.INT,
    help=(
        "Number of articles. The API default is to return 50 articles "
        "matching the provided uid, but the `n` option is used to "
        "customize this value."
    ),
    show_default=True,
)

chunk_size = click.option(
    "-k",
    "--chunksize",
    "chunk_size",
    default=100,
    type=click.INT,
    help=(
        "Size of a simultaneously-processed article chunk. The API default is "
        "to process 100 articles at one time. Decrease this value to "
        "reduce memory usage."
    ),
    show_default=True,
)

concurrency = click.option(
    "-c",
    "--concurrency",
    "concurrency",
    default=5,
    type=click.INT,
    help=(
        "Number of concurrent requests. The API default is to send 5 "
        "concurrent requests, but can be increased to 100. You may want to "
        "limit concurrency to avoid 429 errors on the TrendHunter API."
    ),
    show_default=True,
)

proxy = click.option(
    "-y",
    "--proxy",
    "proxy",
    type=click.STRING,
    help=(
        "The HTTP url of a proxy server. The API default is to not use a "
        "proxy server, but if the TrendHunter API bans your IP address, you "
        "can provide one here. Please try to use a VPN before resorting to "
        "using a proxy server. If you do need to use a proxy, please be aware "
        "of the considerable risk if the provider is not secure."
    ),
)

timeout = click.option(
    "-t",
    "--timeout",
    "timeout",
    default=10,
    type=click.INT,
    help=(
        "Number of seconds until a request times out. The API default is to "
        "allow 10 seconds for a request to complete. If you are receiving "
        "several timeout exceptions, try to increaase this value."
    ),
    show_default=True,
)

format = click.option(
    "-f",
    "--format",
    "format",
    default="0",
    type=click.Choice(["0", "1"]),
    help=(
        "The output format. The API default (0) is to format the output to the"
        " console. If the user would prefer to output the details to a"
        " PowerPoint file, the 1 value can be used."
    ),
    show_default=True,
)

log_level = click.option(
    "-l",
    "--loglevel",
    "log_level",
    default=str(logging.INFO),
    type=click.Choice(
        [
            str(logging.DEBUG),
            str(logging.INFO),
            str(logging.WARNING),
            str(logging.ERROR),
            str(logging.CRITICAL),
        ]
    ),
    help=(
        "The log level of the root Python logger. The API default is to "
        "log anything at or above the INFO level. Decrease the value to "
        "view more verbose logs."
    ),
    show_default=True,
)

path = click.option(
    "-p",
    "--path",
    "path",
    type=click.Path(
        file_okay=True,
        dir_okay=True,
        readable=False,
        writable=True,
        resolve_path=True,
        path_type=Path,
    ),
    help=(
        "The path to write any output files. If one is not "
        "passed, the output path will be the current path."
    ),
)

pixels = click.option(
    "-x",
    "--pixels",
    "pixels",
    default=(300, 300),
    type=click.Tuple([int, int]),
    nargs=2,
    help=(
        "The maximum resolution of any created image files. The API "
        "default is to limit a thumbnail to a dimension of (300, 300). "
        "If an image is not equal in width and height "
        "dimension, the increase in resolution will be halted when "
        "the aspect ratio forces the larger dimension to hit "
        "the boundary specified here."
    ),
    show_default=True,
)

uid = click.argument("uid")


@click.group()
def cli() -> None:
    pass


@cli.command()
@n
@chunk_size
@concurrency
@proxy
@timeout
@format
@log_level
@path
@pixels
@uid
def search(
    n: int,
    chunk_size: int,
    concurrency: int,
    proxy: Optional[str],
    timeout: int,
    format: str,
    log_level: str,
    path: Optional[Path],
    pixels: Tuple[int],
    uid: str,
):
    logging.basicConfig(level=int(log_level))
    formatters = set([FORMATTERS[0], FORMATTERS[int(format)]])
    context = Context(slugify(uid), PageType.SEARCH, pixels, path)

    with TrendHunterAPI(n, chunk_size, concurrency, proxy, timeout) as api:
        for articles in catch_execute(
            api.execute,
            context.uid,
            context.page_type,
        ):
            for formatter in formatters:
                formatter(articles, context)


@cli.command()
@n
@chunk_size
@concurrency
@proxy
@timeout
@format
@log_level
@path
@pixels
@uid
def categories(
    n: int,
    chunk_size: int,
    concurrency: int,
    proxy: Optional[str],
    timeout: int,
    format: str,
    log_level: str,
    path: Optional[Path],
    pixels: Tuple[int],
    uid: str,
):
    logging.basicConfig(level=int(log_level))
    formatters = set([FORMATTERS[0], FORMATTERS[int(format)]])
    context = Context(
        slugify(uid, separator=""), PageType.CATAGORY, pixels, path
    )

    with TrendHunterAPI(n, chunk_size, concurrency, proxy, timeout) as api:
        for articles in catch_execute(
            api.execute, context.uid, context.page_type
        ):
            for formatter in formatters:
                formatter(articles, context)


@cli.command()
@n
@chunk_size
@concurrency
@proxy
@timeout
@format
@log_level
@path
@pixels
@uid
def lists(
    n: int,
    chunk_size: int,
    concurrency: int,
    proxy: Optional[str],
    timeout: int,
    format: str,
    log_level: str,
    path: Optional[Path],
    pixels: Tuple[int],
    uid: str,
):
    logging.basicConfig(level=int(log_level))
    formatters = set([FORMATTERS[0], FORMATTERS[int(format)]])
    context = Context(slugify(uid), PageType.LIST, pixels, path)

    with TrendHunterAPI(n, chunk_size, concurrency, proxy, timeout) as api:
        for articles in catch_execute(
            api.execute, context.uid, context.page_type
        ):
            for formatter in formatters:
                formatter(articles, context)


@cli.command()
@n
@chunk_size
@concurrency
@proxy
@timeout
@format
@log_level
@path
@pixels
@uid
def trends(
    n: int,
    chunk_size: int,
    concurrency: int,
    proxy: Optional[str],
    timeout: int,
    format: str,
    log_level: str,
    path: Optional[Path],
    pixels: Tuple[int],
    uid: str,
):
    logging.basicConfig(level=int(log_level))
    formatters = set([FORMATTERS[0], FORMATTERS[int(format)]])
    context = Context(slugify(uid), PageType.TREND, pixels, path)

    with TrendHunterAPI(n, chunk_size, concurrency, proxy, timeout) as api:
        for articles in catch_execute(
            api.execute, context.uid, context.page_type
        ):
            for formatter in formatters:
                formatter(articles, context)



"""Core API code."""

import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Semaphore
from itertools import chain
from typing import Dict, Generator, List, Optional, Tuple, Union
from urllib.parse import urlencode, urljoin, urlparse, urlunparse

from aiohttp import ClientError, ClientSession
from bs4 import BeautifulSoup

from trendhunter.taxonomies import AJAXIncrement, APIAction, PageType
from trendhunter.tuples import (
    Article,
    Image,
    Metadata,
    Resource,
    Text,
    URLPair,
)
from trendhunter.utils import format_console

TRENDHUNTER_URL = "https://www.trendhunter.com/"


class ResourceError(Exception):
    pass


def sync_url_and_params(url: str, params: Dict) -> str:
    parts = list(urlparse(url))
    parts[4] = urlencode(params or {})
    return urlunparse(parts)


async def _async_get_url(
    url: str,
    semaphore: Semaphore,
    session: ClientSession,
    proxy: Optional[str],
    timeout: int,
    propagate: bool,
) -> Optional[Resource]:
    async with semaphore:
        try:
            format_console(__name__).info(
                f'Downloading content from "{url}"...'
            )

            async with session.get(
                url,
                proxy=proxy,
                allow_redirects=False,
                timeout=timeout,
                raise_for_status=True,
            ) as resp:
                content = await resp.read()

            if not (content.find(b'{"success":true,"data":""}') == -1):
                raise ResourceError(f'No content retrieved from "{url}".')

            return Resource(url, content)
        except (ClientError, ResourceError) as e:
            if propagate:
                raise e
            else:
                format_console(__name__).error(
                    f'HTTP error downloading content of "{url}". Error: {e}.'
                )


async def _async_bridge(
    urls: Union[List[str], Tuple[str], str],
    semaphore: Semaphore,
    session: ClientSession,
    proxy: Optional[str],
    timeout: int,
    propagate: bool,
) -> Union[List[Resource], Resource]:
    if isinstance(urls, (list, tuple)):
        resources = []

        for url in urls:
            resources.append(
                await _async_get_url(
                    url, semaphore, session, proxy, timeout, propagate
                )
            )

        return resources
    else:
        return await _async_get_url(
            urls, semaphore, session, proxy, timeout, propagate
        )


async def async_get_urls(
    urls: Union[List[List[str]], List[Tuple[str]], List[str]],
    concurrency: int,
    proxy: Optional[str],
    timeout: int,
    propagate: bool = False,
) -> Union[List[List[Resource]], List[Resource]]:
    semaphore = Semaphore(concurrency)

    async with ClientSession() as session:
        tasks = [
            _async_bridge(url, semaphore, session, proxy, timeout, propagate)
            for url in urls
        ]
        return await asyncio.gather(*tasks)


def extract_urls(resource: Resource) -> List[URLPair]:
    soup = BeautifulSoup(resource.content, "html.parser")

    def href_contains_trends(href):
        return not (href.replace("\\", "").find("/trends/") == -1)

    urls = []
    elements = soup.find_all(
        "a", attrs={"class": "thar", "href": href_contains_trends}
    )

    for element in elements:
        image = element.find("img", attrs={"data-src": True})

        if image:
            urls.append(
                URLPair(
                    urljoin(
                        TRENDHUNTER_URL, element["href"].replace("\\", "")
                    ),
                    image["data-src"].replace("\\", ""),
                )
            )

    return urls


def extract_image_url(resource: Resource) -> Image:
    soup = BeautifulSoup(resource.content, "html.parser")
    img = soup.find("img", attrs={"class": "gal__mainImage", "data-src": True})
    image_url = img["data-src"].replace("\\", "") if img else None
    return Image(image_url)


def extract_text(resource: Resource) -> Text:
    soup = BeautifulSoup(resource.content, "html.parser")
    title = soup.find("h2", attrs={"class": "tha__title2"})
    description = soup.find("div", attrs={"class": "tha__articleText"})
    meta = soup.find(
        "div",
        attrs={
            "class": "th__article",
            "data-eid": True,
            "data-cid": True,
        },
    )

    title_text = title.get_text().strip() if title else None
    description_text = description.get_text().strip() if description else None
    metadata = Metadata(meta["data-eid"], meta["data-cid"]) if meta else None

    return Text(title_text, description_text, metadata)


class URLIterator(ABC):
    def __init__(self):
        self._increment = 0
        self.increment = 0

    def __iter__(self):
        self.increment = self._increment
        return self

    @abstractmethod
    def __next__(self):
        pass


class ArticleURLIterator(URLIterator):
    def __init__(
        self,
        eid: str,
        cid: str,
        page_type: PageType,
    ):
        super().__init__()
        self.eid = eid
        self.cid = cid
        self.page_type = page_type

    def __next__(self) -> str:
        """"""
        self.increment += 1

        params = {
            "act": APIAction.LP,
            "p": self.increment,
            "aj": AJAXIncrement.ONE,
        }

        if self.increment == 1:
            params["eid"] = self.eid
        else:
            params["cid"] = self.cid

        return sync_url_and_params(
            urljoin(TRENDHUNTER_URL, self.page_type), params
        )


class PageTypeURLIterator(URLIterator):
    def __init__(self, uid: str, page_type: PageType):
        super().__init__()
        self.uid = uid
        self.page_type = page_type

    def __next__(self) -> str:
        """"""
        # page type iterator begins at 1
        self.increment += 1

        params = {
            "act": APIAction.LP,
            "p": self.increment,
            "aj": AJAXIncrement.ONE,
            "pt": self.page_type,
            "v": self.uid,
            "t": PageType.TREND,
        }

        return sync_url_and_params(TRENDHUNTER_URL, params)


class TrendHunterAPI:
    def __init__(
        self,
        n: int,
        chunk_size: int,
        concurrency: int,
        proxy: Optional[str],
        timeout: int,
    ):
        self.n = n
        self.chunk_size = chunk_size
        self.concurrency = concurrency
        self.proxy = proxy
        self.timeout = timeout

    def __enter__(self):
        logging.getLogger("asyncio").setLevel(logging.CRITICAL + 1)
        logging.getLogger("PIL.PngImagePlugin").setLevel(logging.CRITICAL + 1)

        return self

    def __exit__(self, type, value, traceback):
        pass

    def _fetch_all(
        self,
        iterator: URLIterator,
        n: Optional[int] = None,
        extra: List[List[str]] = [],
        extra_resources: List[Tuple[str]] = [],
    ) -> Generator[Article, None, None]:
        all_urls = []
        all_unique_urls = set()
        failsafe = 0
        iterator = chain(iter(extra_resources), iterator)

        while len(all_urls) < (n if n is not None else self.n):
            # fetch 3 results from the iterator at a time
            resource_urls = [next(iterator), next(iterator), next(iterator)]
            resources = asyncio.run(
                async_get_urls(
                    resource_urls,
                    self.concurrency,
                    self.proxy,
                    self.timeout,
                    propagate=True,
                )
            )

            for resource in filter(None, resources):
                limit = (n if n is not None else self.n) - len(all_urls)
                extracted_urls = extract_urls(resource)[:limit]

                for urls in extracted_urls:
                    # do not wrap in `URLPair`, would just have to
                    # be unpacked in a moment
                    url_pair = (urls.url, urls.image_url)

                    if url_pair not in all_unique_urls:
                        all_urls.append(list(url_pair))
                        all_unique_urls.add(url_pair)

            failsafe += 1

            if failsafe > 1000:
                raise ResourceError(
                    "Unable to parse out the required resources after "
                    "hitting 1000 resource pages, the failsafe mechanism to "
                    "halt the fetch has been activated."
                )

        # mix in the extra article urls and remove duplicate
        # resourced articles while maintaining order
        all_urls = extra + all_urls

        while all_urls:
            chunk_urls, all_urls = (
                all_urls[: self.chunk_size],
                all_urls[self.chunk_size :],
            )
            subresources = asyncio.run(
                async_get_urls(
                    chunk_urls, self.concurrency, self.proxy, self.timeout
                )
            )

            articles = []

            for urls, resources in zip(chunk_urls, subresources):
                # `_async_get_url` has previously logged this error
                if not (resources[0] and resources[1]):
                    continue

                text = extract_text(resources[0])
                image = Image(urls[1], resources[1].content)

                if not (text.title or text.description or text.metadata):
                    format_console(__name__).error(
                        "The title, description, or metadata was not "
                        f'parsed from "{resources[0].url}".'
                    )
                    continue

                articles.append(Article(urls[0], text, image))

            yield articles

    def _fetch_one(self, url: str) -> Optional[Text]:
        resources = asyncio.run(
            async_get_urls(
                [url],
                self.concurrency,
                self.proxy,
                self.timeout,
                propagate=True,
            )
        )

        return Article(
            url, extract_text(resources[0]), extract_image_url(resources[0])
        )

    def execute(self, uid: str, page_type: PageType):
        extra, extra_resources = [], []

        if page_type in [PageType.TREND, PageType.LIST]:
            article = self._fetch_one(
                urljoin(TRENDHUNTER_URL, f"{page_type}/{uid}")
            )

            if not article.text.metadata:
                raise ResourceError(
                    f'The metadata was not parsed from "{article.url}".'
                )

            if page_type == PageType.TREND and article.image.url:
                extra = [[article.url, article.image.url]]

            extra_resources = [article.url]

            iterator = ArticleURLIterator(
                article.text.metadata.eid,
                article.text.metadata.cid,
                page_type,
            )
        else:
            iterator = PageTypeURLIterator(uid, page_type)

        yield from self._fetch_all(
            iterator, self.n - len(extra), extra, extra_resources
        )



"""Taxonomies."""


class PageType:
    CATAGORY = "cat"
    SEARCH = "search"
    TREND = "trends"
    LIST = "slideshow"


class APIAction:
    LP = "lp"


class AJAXIncrement:
    ONE = 1
    ZERO = 0



"""Tuples."""

from pathlib import Path
from typing import NamedTuple, Optional, Tuple

from trendhunter.taxonomies import PageType


class Image(NamedTuple):

    url: Optional[str] = None
    image: Optional[bytes] = None


class Metadata(NamedTuple):
    """Article id and category id."""

    eid: str
    cid: str


class Text(NamedTuple):

    title: Optional[str] = None
    description: Optional[str] = None
    metadata: Optional[Metadata] = None


class Article(NamedTuple):
    """Article title, description, metadata, and image."""

    url: str
    text: Text
    image: Image


class URLPair(NamedTuple):

    url: str
    image_url: str


class Resource(NamedTuple):
    """HTML resource, including text and url."""

    url: str
    content: str


class Context(NamedTuple):

    uid: str
    page_type: PageType
    pixels: Tuple[int]
    path: Optional[Path] = None
